@@ -36,6 +36,9 @@ def ds_tag(tag):
 def dsig11_tag(tag):
     return "{" + namespaces.dsig11 + "}" + tag
 
+def ec_tag(tag):
+    return "{" + namespaces.ec + "}" + tag
+
 def _remove_sig(signature, idempotent=False):
     """
     Remove the signature node from its parent, keeping any tail element.
@@ -153,7 +156,7 @@ class XMLSignatureProcessor(XMLProcessor):
     def _get_digest(self, data, digest_algorithm):
         hasher = Hash(algorithm=digest_algorithm, backend=default_backend())
         hasher.update(data)
-        return ensure_str(b64encode(hasher.finalize()))
+        return hasher.finalize()
 
     def _get_digest_method(self, digest_algorithm_id, methods=None):
         if methods is None:
@@ -281,7 +284,8 @@ class XMLSigner(XMLSignatureProcessor):
         self._parser = None
 
     def sign(self, data, key=None, passphrase=None, cert=None, reference_uri=None, key_name=None, key_info=None,
-             id_attribute=None, always_add_key_value=False):
+             id_attribute=None, always_add_key_value=False, payload_inclusive_ns_prefixes=frozenset(),
+             signature_inclusive_ns_prefixes=frozenset()):
         """
         Sign the data and return the root element of the resulting XML tree.
 
@@ -328,6 +332,14 @@ class XMLSigner(XMLSignatureProcessor):
             KeyValue or make sure it matches what's in the certificate. This parameter is provided for compatibility
             purposes only.
         :type always_add_key_value: boolean
+        :param payload_inclusive_ns_prefixes:
+            Provide a list of XML namespace prefixes whose declarations should be preserved when canonicalizing the
+            content referenced by the signature (**InclusiveNamespaces PrefixList**).
+        :type inclusive_ns_prefixes: string
+        :param signature_inclusive_ns_prefixes:
+            Provide a list of XML namespace prefixes whose declarations should be preserved when canonicalizing the
+            signature itself (**InclusiveNamespaces PrefixList**).
+        :type inclusive_ns_prefixes: string
 
         :returns:
             A :py:class:`lxml.etree.Element` object representing the root of the XML tree containing the signature and
@@ -352,12 +364,14 @@ class XMLSigner(XMLSignatureProcessor):
             reference_uris = reference_uri
 
         sig_root, doc_root, c14n_inputs, reference_uris = self._unpack(data, reference_uris)
-        signed_info_element, signature_value_element = self._build_sig(sig_root, reference_uris, c14n_inputs)
-
+        signed_info_element, signature_value_element = self._build_sig(sig_root, reference_uris, c14n_inputs,
+                                                                       sig_insp=signature_inclusive_ns_prefixes,
+                                                                       payload_insp=payload_inclusive_ns_prefixes)
         if key is None:
             raise InvalidInput('Parameter "key" is required')
 
-        signed_info_c14n = self._c14n(signed_info_element, algorithm=self.c14n_alg)
+        signed_info_c14n = self._c14n(signed_info_element, algorithm=self.c14n_alg,
+                                      inclusive_ns_prefixes=signature_inclusive_ns_prefixes)
         if self.sign_alg.startswith("hmac-"):
             from cryptography.hazmat.primitives.hmac import HMAC
             signer = HMAC(key=key,
@@ -475,9 +489,12 @@ class XMLSigner(XMLSignatureProcessor):
             reference_uris = ["#object"]
         return sig_root, doc_root, c14n_inputs, reference_uris
 
-    def _build_sig(self, sig_root, reference_uris, c14n_inputs):
+    def _build_sig(self, sig_root, reference_uris, c14n_inputs, sig_insp, payload_insp):
         signed_info = SubElement(sig_root, ds_tag("SignedInfo"), nsmap=self.namespaces)
-        SubElement(signed_info, ds_tag("CanonicalizationMethod"), Algorithm=self.c14n_alg)
+        sig_c14n_method = SubElement(signed_info, ds_tag("CanonicalizationMethod"), Algorithm=self.c14n_alg)
+        if sig_insp:
+            SubElement(sig_c14n_method, ec_tag("InclusiveNamespaces"), PrefixList=" ".join(sig_insp))
+
         if self.sign_alg.startswith("hmac-"):
             algorithm_id = self.known_hmac_digest_tags[self.sign_alg]
         else:
@@ -490,13 +507,15 @@ class XMLSigner(XMLSignatureProcessor):
                 SubElement(transforms, ds_tag("Transform"), Algorithm=namespaces.ds + "enveloped-signature")
                 SubElement(transforms, ds_tag("Transform"), Algorithm=self.c14n_alg)
             else:
-                SubElement(transforms, ds_tag("Transform"), Algorithm=self.c14n_alg)
+                c14n_xform = SubElement(transforms, ds_tag("Transform"), Algorithm=self.c14n_alg)
+                if payload_insp:
+                    SubElement(c14n_xform, ec_tag("InclusiveNamespaces"), PrefixList=" ".join(payload_insp))
 
             SubElement(reference, ds_tag("DigestMethod"), Algorithm=self.known_digest_tags[self.digest_alg])
             digest_value = SubElement(reference, ds_tag("DigestValue"))
-            payload_c14n = self._c14n(c14n_inputs[i], algorithm=self.c14n_alg)
+            payload_c14n = self._c14n(c14n_inputs[i], algorithm=self.c14n_alg, inclusive_ns_prefixes=payload_insp)
             digest = self._get_digest(payload_c14n, self._get_digest_method_by_tag(self.digest_alg))
-            digest_value.text = digest
+            digest_value.text = ensure_str(b64encode(digest))
         signature_value = SubElement(sig_root, ds_tag("SignatureValue"))
         return signed_info, signature_value
 
@@ -547,7 +566,7 @@ class XMLVerifier(XMLSignatureProcessor):
         if der_encoded_key_value is not None:
             key = load_der_public_key(b64decode(der_encoded_key_value.text), backend=default_backend())
         if "ecdsa-" in signature_alg:
-            if key_value:
+            if key_value is not None:
                 ec_key_value = self._find(key_value, "ECKeyValue", namespace="dsig11")
                 named_curve = self._find(ec_key_value, "NamedCurve", namespace="dsig11")
                 public_key = self._find(ec_key_value, "PublicKey", namespace="dsig11")
@@ -567,7 +586,7 @@ class XMLVerifier(XMLSignatureProcessor):
                 ),
             )
         elif "dsa-" in signature_alg:
-            if key_value:
+            if key_value is not None:
                 dsa_key_value = self._find(key_value, "DSAKeyValue")
                 p = self._get_long(dsa_key_value, "P")
                 q = self._get_long(dsa_key_value, "Q")
@@ -583,7 +602,7 @@ class XMLVerifier(XMLSignatureProcessor):
                        data=signed_info_c14n,
                        algorithm=self._get_signature_digest_method(signature_alg))
         elif "rsa-" in signature_alg:
-            if key_value:
+            if key_value is not None:
                 rsa_key_value = self._find(key_value, "RSAKeyValue")
                 modulus = self._get_long(rsa_key_value, "Modulus")
                 exponent = self._get_long(rsa_key_value, "Exponent")
@@ -641,9 +660,9 @@ class XMLVerifier(XMLSignatureProcessor):
 
         return payload
 
-    def verify(self, data, require_x509=True, x509_cert=None, cert_subject_name=None, ca_pem_file=None, ca_path=None,
-               hmac_key=None, validate_schema=True, parser=None, uri_resolver=None, id_attribute=None,
-               expect_references=1, ignore_ambiguous_key_info=False):
+    def verify(self, data, require_x509=True, x509_cert=None, cert_subject_name=None, cert_resolver=None,
+               ca_pem_file=None, ca_path=None, hmac_key=None, validate_schema=True, parser=None, uri_resolver=None,
+               id_attribute=None, expect_references=1, ignore_ambiguous_key_info=False):
         """
         Verify the XML signature supplied in the data and return the XML node signed by the signature, or raise an
         exception if the signature is not valid. By default, this requires the signature to be generated using a valid
@@ -685,6 +704,16 @@ class XMLVerifier(XMLSignatureProcessor):
             ``None``, requires that the signature supply a valid X.509 certificate chain that validates against the
             known certificate authorities. Implies **require_x509=True**.
         :type x509_cert: string or OpenSSL.crypto.X509
+        :param cert_subject_name:
+            Subject Common Name to check the signing X.509 certificate against. Implies **require_x509=True**.
+        :type cert_subject_name: string
+        :param cert_resolver:
+            Function to use to resolve trusted X.509 certificates when X509IssuerSerial and X509Digest references are
+            found in the signature. The function is called with the keyword arguments ``x509_issuer_name``,
+            ``x509_serial_number`` and ``x509_digest``, and is expected to return an iterable of one or more
+            strings containing a PEM-formatted certificate and a chain of intermediate certificates, if needed.
+            Implies **require_x509=True**.
+        :type cert_resolver: callable
         :param ca_pem_file:
             Filename of a PEM file containing certificate authority information to use when verifying certificate-based
             signatures.
@@ -694,9 +723,6 @@ class XMLVerifier(XMLSignatureProcessor):
             certificate-based signatures. If neither **ca_pem_file** nor **ca_path** is given, the Mozilla CA bundle
             provided by :py:mod:`certifi` will be loaded.
         :type ca_path: string
-        :param cert_subject_name:
-            Subject Common Name to check the signing X.509 certificate against. Implies **require_x509=True**.
-        :type cert_subject_name: string
         :param hmac_key: If using HMAC, a string containing the shared secret.
         :type hmac_key: string
         :param validate_schema: Whether to validate **data** against the XML Signature schema.
@@ -705,7 +731,9 @@ class XMLVerifier(XMLSignatureProcessor):
             Custom XML parser instance to use when parsing **data**. The default parser arguments used by SignXML are:
             ``resolve_entities=False``. See https://lxml.de/FAQ.html#how-do-i-use-lxml-safely-as-a-web-service-endpoint.
         :type parser: :py:class:`lxml.etree.XMLParser` compatible parser
-        :param uri_resolver: Function to use to resolve reference URIs that don't start with "#".
+        :param uri_resolver:
+            Function to use to resolve reference URIs that don't start with "#". The function is called with a single
+            string argument containing the URI to be resolved, and is expected to return a lxml.etree node or string.
         :type uri_resolver: callable
         :param id_attribute:
             Name of the attribute whose value ``URI`` refers to. By default, SignXML will search for "Id", then "ID".
@@ -734,7 +762,7 @@ class XMLVerifier(XMLSignatureProcessor):
         self.x509_cert = x509_cert
         self._parser = parser
 
-        if x509_cert:
+        if x509_cert or cert_resolver:
             self.require_x509 = True
 
         if id_attribute is not None:
@@ -772,11 +800,24 @@ class XMLVerifier(XMLSignatureProcessor):
                 if x509_data is None:
                     raise InvalidInput("Expected a X.509 certificate based signature")
                 certs = [cert.text for cert in self._findall(x509_data, "X509Certificate")]
-                if not certs:
-                    msg = "Expected to find an X509Certificate element in the signature"
-                    msg += " (X509SubjectName, X509SKI are not supported)"
-                    raise InvalidInput(msg)
-                cert_chain = [load_certificate(FILETYPE_PEM, add_pem_header(cert)) for cert in certs]
+                if len(certs) == 0:
+                    x509_iss = x509_data.find("ds:X509IssuerSerial/ds:X509IssuerName", namespaces=namespaces)
+                    x509_sn = x509_data.find("ds:X509IssuerSerial/ds:X509SerialNumber", namespaces=namespaces)
+                    x509_digest = x509_data.find("dsig11:X509Digest", namespaces=namespaces)
+                    if cert_resolver and any(i is not None for i in (x509_iss, x509_sn, x509_digest)):
+                        cert_chain = cert_resolver(x509_issuer_name=x509_iss.text if x509_iss is not None else None,
+                                                   x509_serial_number=x509_sn.text if x509_sn is not None else None,
+                                                   x509_digest=x509_digest.text if x509_digest is not None else None)
+                        if len(cert_chain) == 0:
+                            raise InvalidCertificate("No certificate found for given X509 data")
+                        if not all(isinstance(c, X509) for c in cert_chain):
+                            cert_chain = [load_certificate(FILETYPE_PEM, add_pem_header(cert)) for cert in cert_chain]
+                    else:
+                        msg = "Expected to find an X509Certificate element in the signature"
+                        msg += " (X509SubjectName, X509SKI are not supported)"
+                        raise InvalidInput(msg)
+                else:
+                    cert_chain = [load_certificate(FILETYPE_PEM, add_pem_header(cert)) for cert in certs]
                 signing_cert = verify_x509_cert_chain(cert_chain, ca_pem_file=ca_pem_file, ca_path=ca_path)
             elif isinstance(self.x509_cert, X509):
                 signing_cert = self.x509_cert
@@ -834,11 +875,11 @@ class XMLVerifier(XMLSignatureProcessor):
             copied_root = self.fromstring(self.tostring(root))
             copied_signature_ref = self._get_signature(copied_root)
             transforms = self._find(reference, "Transforms", require=False)
-            digest_algorithm = self._find(reference, "DigestMethod").get("Algorithm")
+            digest_alg = self._find(reference, "DigestMethod").get("Algorithm")
             digest_value = self._find(reference, "DigestValue")
             payload = self._resolve_reference(copied_root, reference, uri_resolver=uri_resolver)
             payload_c14n = self._apply_transforms(payload, transforms, copied_signature_ref, c14n_algorithm)
-            if digest_value.text != self._get_digest(payload_c14n, self._get_digest_method(digest_algorithm)):
+            if b64decode(digest_value.text) != self._get_digest(payload_c14n, self._get_digest_method(digest_alg)):
                 raise InvalidDigest("Digest mismatch for reference {}".format(len(verify_results)))
 
             # We return the signed XML (and only that) to ensure no access to unsigned data happens
